% This Matlab file implements Doodle Segment Querying and Colorization by
% Joe Kohlmann and Nathalie Cheng. This project uses code adapted from the
% following original authors:
% 
% 1. FMIQ implementation in Python by Ajit Datar:
%    http://www.d.umn.edu/~data0003/Projects/FMIQ/
% 2. Colorization algorithm by Anat Levin, Dani Lischinski, Yair Weiss:
%    http://www.cs.huji.ac.il/~yweiss/Colorization/
% 3. imoverlay.m by Steve Eddins of MathWorks:
%	 http://www.mathworks.com/matlabcentral/fileexchange/10502
% 
% We claim no rights to the original authors' works. We redistribute this
% code as part of our CS 534 final project for academic purposes only.
% 
% Please note that the FMIQ algorithm requires the Numeric and PIL (Python
% Imaging Library) modules. Obtain these from the following sources:
% Numeric: http://sourceforge.net/projects/numpy/files/Old%20Numeric/
% PIL: http://www.pythonware.com/products/pil/
% My test system was a Mac running Mac OS X 10.6.7 and Python 2.4.

function [ outputImage ] = ComposeDoodle(doodleColorName, trueColorName, originalPath, tempPath, ignoreHexColor)
    % Defaults
    if nargin < 1, doodleColorName = 'CollageDoodleColor.png'; end;
    if nargin < 2, trueColorName   = 'CollageTrueColor.png'; end;
    if nargin < 3, originalPath    = 'originals'; end;
    if nargin < 4, tempPath        = 'temp'; end;
    if nargin < 5, ignoreHexColor  = '#FFFFFF'; end;
    
    % Path Modification
    oldP = path;
    newP = genpath( [pwd, '/colorization/'] );
    path( path, newP );
    newP = genpath( [pwd, '/fmiq/'] );
    path( path, newP );
    newP = genpath( [pwd, '/imoverlay/'] );
    path( path, newP );
    
    % Python Path
    % Change this to correspond to a Python install on your system that
    % includes Numeric and the Python Imaging Library (PIL).
    pythonPath = '/opt/local/bin/python';
    
    % Argument Checks
    % Make sure doodleColorName references a PNG.
    if ~strcmpi( doodleColorName((end - 3):end) , '.png')
        % Error: Doodle image must be a PNG.
        theException = MException('ComposeDoodle:BadArgument', 'doodleColorName does not reference a PNG.');
        throw(theException)
    end
    
    % Make sure trueColorName references a PNG.
    if ~strcmpi( trueColorName((end - 3):end) , '.png')
        % Error: True color image must be a PNG.
        theException = MException('ComposeDoodle:BadArgument', 'trueColorName does not reference a PNG.');
        throw(theException)
    end
    
    % Make sure originalPath references an existing folder.
    if exist(originalPath,'file') ~= 7
        % Error: originalPath doesn't exist.
        theException = MException('ComposeDoodle:BadArgument', 'originalPath does not reference an existing folder.');
        throw(theException)
    end
    
    % Make sure tempPath references an existing folder.
    if exist(tempPath,'file') ~= 7
        % Error: tempPath doesn't exist.
        theException = MException('ComposeDoodle:BadArgument', 'tempPath does not reference an existing folder.');
        throw(theException)
    end
    
    % Make sure ignoreHexColor is an 8-bit RGB hex value (HTML/CSS-style).
    if ~strcmp(ignoreHexColor(1),'#') || length(ignoreHexColor) ~= 7
        % Error: ignoreHexColor is in the wrong format.
        theException = MException('ComposeDoodle:BadArgument', 'ignoreHexColor is not an 8-bit hexadecimal color value.');
        throw(theException)
    end
    
    %%%% FMIQ Preprocessing %%%%
    % Tell the Python FMIQ code to add the directory of original images to
    % its database (if the database doesn't already exist).
    if exist('fmiq/db_imagedb','file') ~= 2
        system(['cd ./fmiq/ && ' pythonPath ' adddir.py --dir ../' originalPath]);
    end
    
    %%%% Doodle Segment Separation %%%%
    % Load the doodle color image and the true color image.
    % Convert them to 8-bit RGB.
    dI = im2uint8(imread(doodleColorName));
    cI = im2uint8(imread(trueColorName));
    [dIHeight, dIWidth, dIChannels] = size(dI);
    [cIHeight, cIWidth, cIChannels] = size(cI);
    
    if size(dI) ~= size(cI)
        % Error: Doodle color image and true color image must have the same
        % dimensions.
        theException = MException('ComposeDoodle:ImagesDifferentSizes', 'The doodle color image and the true color image have different dimensions.');
        throw(theException)
    end
    
    % Create a new blank image for the output.
    outputImage = repmat(255, [dIHeight, dIWidth, dIChannels]);
    
    % Find unique colors in the image.
    % Adapted from:
    %     http://blogs.mathworks.com/steve/2008/01/31/counting-occurrences-of-image-colors/
    dIColumns = reshape(dI, [], dIChannels);
    [uniqueColors, m, n] = unique(dIColumns, 'rows');
    
    for index = 1:size(uniqueColors)
        indexHexColor = upper(sprintf('#%02x%02x%02x', ...
            uniqueColors(index,1), ...
            uniqueColors(index,2), ...
            uniqueColors(index,3)  ...
        ));
        
        indexColor = [ uniqueColors(index,1) uniqueColors(index,2) uniqueColors(index,3) ];
        
        if strcmp(indexHexColor, ignoreHexColor)
            continue
        end
        % Generate a masked image containing only pixels with this color.
        mask = n == index;
        mask = reshape(mask, dIHeight, dIWidth);
        % Figure out where to crop the image.
        cropCoords = CropBoxFromMask(mask);
        % Save the coordinate of the upper-left part of this segment.
        dIAnchor = [cropCoords(1) cropCoords(2)];
        cropHeight = cropCoords(3);
        cropWidth = cropCoords(4);
        % Create a masked image that shows only this color and white.
        % doodleSegment = imoverlay(dI, ~mask, [1 1 1]);
        % Crop the doodle segment and the true color segment images.
        % doodleSegment = imcrop(doodleSegment, cropCoords);
        trueSegment = imcrop(cI, cropCoords);
        % Crop the mask too.
        cropMask = imcrop(mask, cropCoords);
        % Write the trueSegment to disk as a JPEG for FMIQ.
        trueSegmentPath = [tempPath '/' indexHexColor '.jpg'];
        imwrite(trueSegment, trueSegmentPath, 'jpg', 'Quality', 100);
        
        % Perform FMIQ on the true color segment.
        [status, bestMatch] = system(['cd ./fmiq/ && ' pythonPath ' queryfile.py ../' tempPath '/' indexHexColor '.jpg']);
        bestMatch = strtrim(regexprep(bestMatch, ['../' originalPath '/'], ''));
        
        % Colorize the matched image.
        matchImage = im2uint8(imread([originalPath '/' bestMatch]));
        % Resize the matched image to match the vertical height of the crop.
        matchImage = imresize(matchImage, [cropHeight cropWidth]);
        [matchHeight, matchWidth, matchChannels] = size(matchImage);
        markedImage = matchImage;
        
        % "Doodle" on the match image using the doodle pixels.
        for i = 1:matchHeight
            for j = 1:matchWidth
                if i <= size(cropMask,1) && j <= size(cropMask,2)
                    maskPixel = cropMask(i,j);
                else
                    maskPixel = 0;
                end
                if maskPixel == 1
                    markedImage(i,j,:) = indexColor;
                end
            end
        end
        
        imwrite(markedImage, '/Users/kohlmannj/Desktop/Test.png', 'png');
        
        [markedHeight, markedWidth, markedChannels] = size(markedImage);
        
        % Colorize!!!!!
        colorizedImage = colorizeFun(matchImage, markedImage);
        colorizedImage = imresize(colorizedImage, [cropWidth cropHeight]);
        % imshow(colorizedImage);
        % [outputHeight, outputWidth, outputChannels] = size(outputImage);
        % Copy the colorized image into the output image.
        [cropCoords(1) (cropCoords(1) + cropHeight - 1); cropCoords(2) (cropCoords(2) + cropWidth - 1)]
        outputImageSegment = outputImage(cropCoords(1):cropCoords(1) + cropHeight - 1, cropCoords(2):cropCoords(2) + cropWidth - 1, :);
        [outputHeight, outputWidth, outputChannels] = size(outputImageSegment);
        outputImage(cropCoords(2):cropCoords(2) + cropWidth - 1, cropCoords(1):cropCoords(1) + cropHeight - 1, :) = colorizedImage;
    end
    
    % outputImage(1:128, 1:128, :) = 0;
    
    imwrite(outputImage, 'output.jpg', 'jpg', 'Quality', 100);
    
    % Path Restore
    path(oldP);
end